@@ -1,10 +1,9 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 
-__version__ = '0.1.0'
+__version__ = '0.2.0'
 
 import BaseHTTPServer
-import collections
 from cStringIO import StringIO
 from datetime import datetime, timedelta
 import email.message
@@ -12,6 +11,7 @@ import gzip
 import httplib
 import json
 from optparse import OptionParser
+import os.path
 import random
 import re
 import socket
@@ -25,8 +25,45 @@ import urlparse
 
 DEFAULT_PORT = 13085
 
-Request = collections.namedtuple(
-    'Request', ('method', 'path', 'query', 'headers', 'body'))
+class Request(object):
+    
+    """An HTTP request.
+    
+    .. attribute:: method
+       
+       Request method.
+    
+    .. attribute:: path
+       
+       The request path, excluding the query string.
+    
+    .. attribute:: query
+       
+       A dictionary of parameters parsed from the query string.
+       An empty dictionary if none can be parsed.
+    
+    .. attribute:: headers
+       
+       An ``rfc822.Message``-like mapping of request headers.
+    
+    .. attribute:: body
+       
+       Request entity body as a `str` if there is one, otherwise `None`.
+    """
+    
+    def __init__(self, method, path, query, headers, body):
+        self.method = method
+        self.path = path
+        self.query = query
+        self.headers = headers
+        self.body = body
+    
+    def p(self, x):
+        if callable(x):
+            return x(self)
+        else:
+            return x
+
 
 class Response(object):
     
@@ -111,43 +148,74 @@ class Rule(object):
         self._chain = []
         self._final = None
         self._counter = 0
+        self._maybe = []
         self._enable_cors = False
-        self._enable_jsonp = False
         self._allow = None
         self._gzip = False
     
     @Cheat.entry('code')
     def status(self, code):
+        """Set response status to `code`."""
         self._status = code
         return self
     
     @Cheat.entry('name, value, **params', 'params are appended as k=v pairs')
     def header(self, name, value, **params):
+        """Set response header `name` to `value`.
+        
+        If `params` are specified, they are appended as k=v pairs.
+        For example::
+        
+            header('Content-Type', 'text/html', charset='utf-8')
+        
+        produces::
+        
+            Content-Type: text/html; charset="utf-8"
+        """
         self._headers.append(('set', name, value, params))
         return self
     
     @Cheat.entry('name, value, **params', 'added, not replaced')
     def add_header(self, name, value, **params):
+        """Same as :meth:`~Rule.header`, but add the header, not replace it.
+        
+        This can be used to send multiple headers with the same name,
+        such as ``Via`` or ``Set-Cookie``
+        (but for the latter see :meth:`~Rule.cookie`).
+        """
         self._headers.append(('add', name, value, params))
         return self
     
     @Cheat.entry('data')
     def body(self, data):
+        """Set response entity body to `data`."""
         self._body = data
         return self
     
     @Cheat.entry('path')
     def body_file(self, path):
-        return self.body(open(path).read())
+        """Set response entity body to the contents of `path`.
+        
+        The file at `path` is read when the rules are posted.
+        `path` undergoes tilde expansion.
+        """
+        return self.body(open(os.path.expanduser(path)).read())
     
     @Cheat.entry('url', 'data from <var>url</var> is cached until Turq exits')
     def body_url(self, url):
+        """Set response entity body to the contents of `url`.
+        
+        The resource at `url` is fetched once and cached until Turq exits.
+        Note that this method only sets the entity body.
+        HTTP status and headers are not copied from `url`.
+        """
         if url not in Rule.url_cache:
             Rule.url_cache[url] = urllib2.urlopen(url).read()
         return self.body(Rule.url_cache[url])
     
     @Cheat.entry('seconds')
     def delay(self, seconds):
+        """Delay for `seconds` before serving the response."""
         self._delay = seconds
         return self
     
@@ -157,19 +225,31 @@ class Rule(object):
     
     @Cheat.entry('mime_type', 'content type')
     def ctype(self, value):
+        """Set response ``Content-Type`` to `value`."""
         return self.header('Content-Type', value)
     
     @Cheat.entry('[text]', 'plain text')
     def text(self, text='Hello world!'):
+        """Set up a ``text/plain`` response."""
         return self.ctype('text/plain; charset=utf-8').body(text)
     
     @Cheat.entry('[nbytes]', 'roughly <var>nbytes</var> of plain text')
     def lots_of_text(self, nbytes=20000):
+        """Set up a ``text/plain`` response with lots of text.
+        
+        Lines of dummy text will be generated
+        so that the entity body is very roughly `nbytes` in length.
+        """
         return self.text(make_text(nbytes))
     
     @Cheat.entry('[title], [text]', 'basic HTML page')
     def html(self, title='Hello world!', text='This is Turq!'):
-        body = '''<!DOCTYPE html>
+        """Set up a ``text/html`` response.
+        
+        A basic HTML page with `title` and a paragraph of `text` is served.
+        """
+        return self.ctype('text/html; charset=utf-8').body(lambda req: (
+            '''<!DOCTYPE html>
 <html>
     <head>
         <title>%s</title>
@@ -178,11 +258,15 @@ class Rule(object):
         <h1>%s</h1>
         <p>%s</p>
     </body>
-</html>''' % (title, title, text)
-        return self.ctype('text/html; charset=utf-8').body(body)
+</html>''' % (req.p(title), req.p(title), req.p(text))
+        ))
     
     @Cheat.entry('[nbytes]', 'roughly <var>nbytes</var> of HTML')
     def lots_of_html(self, nbytes=20000, title='Hello world!'):
+        """Set up a ``text/html`` response with lots of text.
+        
+        Like :meth:`~Rule.lots_of_text`, but wrapped in HTML paragraphs.
+        """
         return self.html(
             title=title,
             text=make_text(nbytes - 100).replace('\n\n', '</p><p>')
@@ -192,60 +276,114 @@ class Rule(object):
                  'JSONP is handled automatically, '
                  'pass <code>jsonp=False</code> to disable')
     def json(self, data={'result': 'turq'}, jsonp=True):
-        self._enable_jsonp = jsonp
-        return self.ctype('application/json').body(json.dumps(data))
+        """Set up a JSON or JSONP response.
+        
+        `data` will be serialized into an ``application/json`` entity body.
+        But if the request has a ``callback`` query parameter,
+        `data` will be wrapped into a JSONP callback
+        and served as ``application/javascript``,
+        unless you set `jsonp` to `False`.
+        """
+        self.ctype(lambda req: 'application/javascript'
+                               if jsonp and 'callback' in req.query
+                               else 'application/json')
+        self.body(lambda req: (
+            '%s(%s);' % (req.query['callback'], json.dumps(req.p(data)))
+            if jsonp and 'callback' in req.query
+            else json.dumps(req.p(data))
+        ))
+        return self
     
     @Cheat.entry('[code]', 'JavaScript')
     def js(self, code='alert("Turq");'):
+        """Set up an ``application/javascript`` response."""
         return self.ctype('application/javascript').body(code)
     
     @Cheat.entry('[code]')
     def xml(self, code='<turq></turq>'):
+        """Set up an ``application/xml`` response."""
         return self.ctype('application/xml').body(code)
     
     @Cheat.entry('location, [status=302]')
     def redirect(self, location, status=httplib.FOUND):
+        """Set up a redirection response."""
         return self.status(status).header('Location', location)
     
     @Cheat.entry('name, value, [max_age], [path]...')
     def cookie(self, name, value, max_age=None, expires=None, path=None,
                secure=False, http_only=False):
-        data = '%s=%s' % (name, value)
-        if max_age is not None:
-            data += '; Max-Age=%s' % max_age
-        if expires is not None:
-            data += '; Expires=%s' % expires
-        if path is not None:
-            data += '; Path=%s' % path
-        if secure:
-            data += '; Secure'
-        if http_only:
-            data += '; HttpOnly'
-        return self.add_header('Set-Cookie', data) 
+        """Add a cookie `name` with `value`.
+        
+        The other arguments
+        correspond to parameters of the ``Set-Cookie`` header.
+        If specified, `max_age` and `expires` should be strings.
+        Nothing is escaped.
+        """
+        def cookie_string(req):
+            data = '%s=%s' % (req.p(name), req.p(value))
+            if max_age is not None:
+                data += '; Max-Age=%s' % req.p(max_age)
+            if expires is not None:
+                data += '; Expires=%s' % req.p(expires)
+            if path is not None:
+                data += '; Path=%s' % req.p(path)
+            if secure:
+                data += '; Secure'
+            if http_only:
+                data += '; HttpOnly'
+            return data
+        return self.add_header('Set-Cookie', cookie_string)
     
     @Cheat.entry('[realm]')
     def basic_auth(self, realm='Turq'):
-        return self.status(httplib.UNAUTHORIZED). \
-                    header('WWW-Authenticate', 'Basic realm="%s"' % realm)
+        """Demand HTTP basic authentication (status code 401)."""
+        self.status(httplib.UNAUTHORIZED)
+        self.header('WWW-Authenticate',
+                    lambda req: 'Basic realm="%s"' % req.p(realm))
+        return self
     
     @Cheat.entry('[realm], [nonce]')
     def digest_auth(self, realm='Turq', nonce='twasbrillig'):
-        return self.status(httplib.UNAUTHORIZED). \
-                    header('WWW-Authenticate',
-                           'Digest realm="%s", nonce="%s"' % (realm, nonce))
+        """Demand HTTP digest authentication (status code 401)."""
+        self.status(httplib.UNAUTHORIZED)
+        self.header(
+            'WWW-Authenticate',
+            lambda req: 'Digest realm="%s", nonce="%s"' % (
+                req.p(realm), req.p(nonce)
+            )
+        )
+        return self
     
     @Cheat.entry('*methods', 'otherwise send 405 with a text error message')
     def allow(self, *methods):
+        """Check the request method to be one of `methods` (case-insensitive).
+        
+        If it isn’t, send 405 Method Not Allowed
+        with a plain-text error message.
+        """
         self._allow = set(m.lower() for m in methods)
         return self
     
     @Cheat.entry()
     def cors(self):
+        """Enable `CORS <http://www.w3.org/TR/cors/>`_ on the response.
+        
+        Currently this just sets ``Access-Control-Allow-Origin: *``.
+        Preflight requests are not yet handled.
+        """
         self._enable_cors = True
         return self.header('Access-Control-Allow-Origin', '*')
     
     @Cheat.entry('when', '“10 minutes” or “5 h” or “1 day”')
     def expires(self, when):
+        """Set the expiration time of the response.
+        
+        `when` should be a specification
+        of the number of minutes, hours or days,
+        counted from the moment the rules are posted.
+        Supported formats are: “10 min” or “10 minutes” or “5 h” or “5 hours”
+        or “1 d” or “1 day”.
+        """
         n, unit = when.split()
         n = int(n)
         dt = datetime.utcnow()
@@ -261,17 +399,23 @@ class Rule(object):
     
     @Cheat.entry()
     def gzip(self):
+        """Apply ``Content-Encoding: gzip`` to the entity body.
+        
+        ``Accept-Encoding`` of the request is ignored.
+        """
         self._gzip = True
         return self
     
     @Cheat.entry('', 'begin a sub-rule for the first hit...')
     def first(self):
+        """Add a sub-rule that will be applied on a first request."""
         sub = Rule()
         self._chain = [sub]
         return sub
     
     @Cheat.entry('', '...the next hit...')
     def next(self):
+        """Add a sub-rule that will be applied on a subsequent request."""
         assert self._chain, 'next() without first()'
         sub = Rule()
         self._chain.append(sub)
@@ -279,39 +423,51 @@ class Rule(object):
     
     @Cheat.entry('', '...and all subsequent hits')
     def then(self):
+        """Add a sub-rule that will be applied on all subsequent requests."""
         assert self._chain, 'then() without first()'
         self._final = Rule()
         return self._final
     
+    @Cheat.entry('[probability]', 'start a stochastic sub-rule')
+    def maybe(self, probability=0.1):
+        """Add a sub-rule that will be applied with `probability`."""
+        assert probability > 0
+        sub = Rule()
+        self._maybe.append((sub, probability))
+        assert sum(p for s, p in self._maybe) <= 1
+        return sub
+    
+    @Cheat.entry('', 'complement all <code>maybe</code>s')
+    def otherwise(self):
+        """Add a sub-rule that complements all :meth:`~Rule.maybe` rules.
+        
+        This is just a shortcut
+        that adds a `maybe` sub-rule
+        with a probability equal to 1 minus all currently defined `maybe`.
+        Thus, it must come after all `maybe`.
+        """
+        return self.maybe(1 - sum(p for s, p in self._maybe))
+    
     def __enter__(self):
         return self
     
     def __exit__(self, *args):
         return False
     
-    def _set_header(self, resp, name, value):
-        del resp[name]
-        resp[name] = value
-    
-    def apply_normal(self, resp):
-        if self._delay:
-            time.sleep(self._delay)
+    def apply_normal(self, req, resp):
+        if self._delay is not None:
+            time.sleep(req.p(self._delay))
         if self._status is not None:
-            resp.status = self._status
+            resp.status = req.p(self._status)
         if self._body is not None:
-            resp.body = self._body
+            resp.body = req.p(self._body)
     
-    def apply_headers(self, resp):
+    def apply_headers(self, req, resp):
         for op, name, value, params in self._headers:
             if op == 'set':
-                resp.set_header(name, value, **params)
+                resp.set_header(req.p(name), req.p(value), **params)
             elif op == 'add':
-                resp.add_header(name, value, **params)
-    
-    def apply_jsonp(self, req, resp):
-        if self._enable_jsonp and ('callback' in req.query):
-            resp.body = '%s(%s);' % (req.query['callback'][0], resp.body)
-            resp.set_header('Content-Type', 'application/javascript')
+                resp.add_header(req.p(name), req.p(value), **params)
     
     def apply_chain(self, req, resp):
         if self._counter >= len(self._chain):
@@ -323,6 +479,15 @@ class Rule(object):
             self._chain[self._counter].apply(req, resp)
             self._counter += 1
     
+    def apply_maybe(self, req, resp):
+        x = random.random()
+        pos = 0
+        for sub, prob in self._maybe:
+            pos += prob
+            if x < pos:
+                sub.apply(req, resp)
+                return
+    
     def apply_processor(self, req, resp):
         if self._processor is not None:
             sub = Rule()
@@ -336,7 +501,7 @@ class Rule(object):
                 resp.set_header('Content-Type', 'text/plain')
                 resp.body = 'Method %s not allowed here' % req.method
     
-    def apply_gzip(self, resp):
+    def apply_gzip(self, req, resp):
         if self._gzip and resp.body:
             zbuf = StringIO()
             zfile = gzip.GzipFile(fileobj=zbuf, mode='w')
@@ -346,12 +511,12 @@ class Rule(object):
             resp.set_header('Content-Encoding', 'gzip')
     
     def apply(self, req, resp):
-        self.apply_normal(resp)
-        self.apply_headers(resp)
-        self.apply_jsonp(req, resp)
+        self.apply_normal(req, resp)
+        self.apply_headers(req, resp)
         self.apply_chain(req, resp)
+        self.apply_maybe(req, resp)
         self.apply_allow(req, resp)
-        self.apply_gzip(resp)
+        self.apply_gzip(req, resp)
         self.apply_processor(req, resp)
 
 
@@ -431,10 +596,13 @@ CONSOLE_TPL = string.Template('''
             </div>
             <div>
                 <input type="submit" value="Commit" accesskey="s">
-                <span id="okay" style="color: #0E7C00">$okay</span>
+                <span id="okay">$okay</span>
             </div>
         </form>
-        <p><a href="https://github.com/vfaronov/turq">Turq</a> $version</p>
+        <p>
+            <a href="https://github.com/vfaronov/turq">Turq</a> $version ·
+            <a href="https://turq.readthedocs.org/">docs</a>
+        </p>
     </body>
 </html>
 ''')
@@ -472,9 +640,9 @@ class TurqHandler(BaseHTTPServer.BaseHTTPRequestHandler):
                 q = urlparse.parse_qs(q)
             except Exception:
                 q = urlparse.parse_qs('')       # so it's still iterable
-            return q
         else:
-            return urlparse.parse_qs('')
+            q = urlparse.parse_qs('')
+        return dict((k, v[0]) for k, v in q.items())
     
     @property
     def body(self):
